"""
Support functions for parsing command-line arguments and providing
the Topographica command prompt.  Typically called from the
'./topographica' script, but can be called directly if using
Topographica files within a separate Python.

$Id$
"""


from optparse import OptionParser

import sys, __main__, math, os, re

import topo
import param
from param.parameterized import Parameterized
from topo.base.simulation import OptionalSingleton


try:
    # By default, use a non-GUI backend for matplotlib.
    from matplotlib import rcParams
    rcParams['backend']='Agg'
    matplotlib_imported=True
except ImportError:
    matplotlib_imported=False


ipython_shell_interface = None
ipython_prompt_interface = None
try:
    from IPython.frontend.terminal.embed import InteractiveShellEmbed as IPShell
    from IPython.config.loader import Config
    ipython_shell_interface = "InteractiveShellEmbed"
    try:
        from IPython.core.prompts import PromptManager  # pyflakes:ignore (try/except import)
        ipython_prompt_interface = "PromptManager"
    except ImportError:
        pass
except ImportError:
    try:
        # older version?
        from IPython.Shell import IPShell  # pyflakes:ignore (try/except import)
        ipython_shell_interface = "IPython.Shell"
    except ImportError:
        print "Note: IPython is not available; using basic interactive Python prompt instead."



# Startup banner
BANNER = """
Welcome to Topographica!

Type help() for interactive help with python, help(topo) for general
information about Topographica, help(commandname) for info on a
specific command, or topo.about() for info on this release, including
licensing information.
"""


class GlobalParams(Parameterized,OptionalSingleton):
    """
    A Parameterized class providing script-level parameters.
    
    Script-level parameters can be set from the commandline by passing
    via -p, e.g. ./topographica -p retina_density=10

    Within scripts, parameters can be declared by using the add()
    method.


    Example usage in a script:

    from topo.misc.commandline import global_params as p
    p.add(
        retina_density=param.Number(default=24.0,bounds=(0,None),
        inclusive_bounds=(False,True),doc=\"""
        The nominal_density to use for the retina.\"""))
    ...
    topo.sim['Retina']=sheet.GeneratorSheet(
        nominal_density=p.retina_density)


    Further information:

    'context' is usually set to __main__.__dict__ and is used to find
    the value of a parameter as it is add()ed to this object
    (i.e. add() has access to values set via the commandline or in
    scripts).

    Values set via set_in_context() or exec_in_context() (used by -p)
    are tracked: warnings are issued for overwritten values, and
    unused values can be warned about via check_for_unused_names().

    The context is not saved in snapshots, but parameter values are
    saved.
    """
    context = None

    def __new__(cls,*args,**kw):
        return OptionalSingleton.__new__(cls,True)

    def __init__(self,context=None,**params):
        self.context = context or {}
        self.unused_names = set()
        params['name']="global_params"
        super(GlobalParams,self).__init__(**params)

    def __getstate__(self):
        # context is neither saved nor restored
        # (in our current usage, the context of the GlobalParams
        # instance will be set to __main__.__dict__ on startup).
        state = super(GlobalParams,self).__getstate__()
        del state['context']
        return state

    def set_in_context(self,**params):
        """
        Set in self.context all name=val pairs specified in **params,
        tracking new names and warning of any replacements.
        """
        for name,val in params.items():
            if name in self.context:
                self.warning("Replacing previous value of '%s' with '%s'"%(name,val))
            self.context[name]=val
            self.unused_names.add(name)
            
    def exec_in_context(self,arg):
        """
        exec arg in self.context, tracking new names and
        warning of any replacements.
        """
        ## contains elaborate scheme to detect what is specified by
        ## -s, and to warn about any replacement
        current_ids = dict([(k,id(v)) for k,v in self.context.items()])

        exec arg in self.context

        for k,v in self.context.items():
            if k in self.unused_names and id(v)!=current_ids[k]:
                self.warning("Replacing previous value of '%s' with '%s'"%(k,v))

        new_names = set(self.context.keys()).difference(set(current_ids.keys()))
        for k in new_names:
            self.unused_names.add(k)
        
    def check_for_unused_names(self):
        """Warn about any unused names."""
        for s in self.unused_names:
            self.warning("'%s' is unused."%s)

# warns for param that specified with -c (but also if name gets defined in __main__,
# e.g. by default_density=global_params.default_density in a script file
##         for name in self.params():
##             if name in self.context:
##                 self.warning("'%s' still exists in global_params.context"%name)

        # detect duplicate param value that wasn't used (e.g. specified with after script)
        for name,val in self.params().items():
            if name in self.context:
                if self.context[name]!=self.inspect_value(name):
                    self.warning("'%s=%s' is unused."%(name,self.context[name]))
            
    
    def add(self,**kw):
        """
        For each parameter_name=parameter_object specified in kw:
        * adds the parameter_object to this object's class
        * if there is an entry in context that has the same name as the parameter,
          sets the value of the parameter in this object to that value, and then removes
          the name from context
        """        
        for p_name,p_obj in kw.items():
            self._add_parameter(p_name,p_obj)
            if p_name in self.context:
                setattr(self,p_name,self.context[p_name])
                if p_name in self.unused_names:
                    # i.e. remove from __main__ if it was a -p option (but not if -c)
                    del self.context[p_name]  
                    self.unused_names.remove(p_name)
                

global_params=GlobalParams(context=__main__.__dict__)



##### Command-prompt formatting
#    
class IPCommandPromptHandler(object):
    """
    Allows control over IPython's dynamic command prompts.
    """
    _format = ''
    _prompt = ''

    @classmethod
    def set_format(cls,format):
        """
        Set IPython's prompt template to format.
        """
        import __main__
        IP = __main__.__dict__['__IP']
        prompt = getattr(IP.outputcache,cls._prompt)
        prompt.p_template = format
        prompt.set_p_str()        
        cls._format = format

    @classmethod
    def get_format(cls):
        """
        Return the current template.
        """
        return cls._format

    
class CommandPrompt(IPCommandPromptHandler):
    """
    Control over input prompt.

    Several predefined formats are provided, and any of these (or any
    arbitrary string) can be used by calling set_format() with their
    values.

    See the IPython manual for details:
    http://ipython.scipy.org/doc/manual/html/config/index.html

    Examples:
      # Use one of the predefined formats:
      CommandPrompt.set_format(CommandPrompt.basic_format)
      # Just print the command number:
      CommandPrompt.set_format('\# ')
      # Print the command number but don't use color:
      CommandPrompt.set_format('\N ')
      # Print the value of my_var at each prompt:
      CommandPrompt.set_format('{my_var}>>> ')
    """
    _prompt = 'prompt1'
    
    # Predefined alternatives
    basic_format   = 'Topographica>>> '
    if ipython_prompt_interface == "PromptManager":
        simtime_format = 'topo_t{topo.sim.timestr_prop}>>> '
        simtimecmd_format = 'topo_t{topo.sim.timestr_prop}_c\\#>>> '
    else:
        simtime_format = 'topo_t${topo.sim.timestr()}>>> '
        simtimecmd_format = 'topo_t${topo.sim.timestr()}_c\\#>>> '
    
    _format = simtimecmd_format


class CommandPrompt2(IPCommandPromptHandler):
    """
    Control over continuation prompt.

    (See CommandPrompt.)
    """
    _prompt = 'prompt2'
    basic_format = '   .\\D.: '
    _format = basic_format


class OutputPrompt(IPCommandPromptHandler):
    """
    Control over output prompt.

    (See CommandPrompt.)
    """
    _prompt = 'prompt_out'
    basic_format = 'Out[\#]:'
    _format = basic_format

#####



# Use to define global constants
global_constants = {'pi':math.pi}

# Create the topographica parser.
usage = "usage: topographica ([<option>]:[<filename>])*\n\
where any combination of options and Python script filenames will be\n\
processed in order left to right."
topo_parser = OptionParser(usage=usage)


def sim_name_from_filename(filename):
    """
    Set the simulation title from the given filename, if none has been
    set already.
    """
    if topo.sim.name is None:
        topo.sim.name=re.sub('.ty$','',os.path.basename(filename))


def boolean_option_action(option,opt_str,value,parser):
    """Callback function for boolean-valued options that apply to the entire run.""" 
    #print "Processing %s" % (opt_str)
    setattr(parser.values,option.dest,True)


def interactive():
    os.environ['PYTHONINSPECT']='1'

# CB: note that topographica should stay open if an error occurs
# anywhere after a -i (i.e. in a -c command or script)
def i_action(option,opt_str,value,parser):
    """Callback function for the -i option."""
    boolean_option_action(option,opt_str,value,parser)
    interactive()
    
topo_parser.add_option("-i","--interactive",action="callback",callback=i_action,
                       dest="interactive",default=False,
                       help="provide an interactive prompt even if stdin does not appear to be a terminal.")


def v_action(option,opt_str,value,parser):
    """Callback function for the -v option."""
    import param.parameterized
    param.parameterized.min_print_level=param.parameterized.VERBOSE
    print "Enabling verbose message output."
    
topo_parser.add_option("-v","--verbose",action="callback",callback=v_action,dest="verbose",default=False,help="""\
enable verbose messaging output.""")


def d_action(option,opt_str,value,parser):
    """Callback function for the -d option."""
    import param.parameterized
    param.parameterized.min_print_level=param.parameterized.DEBUG
    print "Enabling debugging message output."
    
topo_parser.add_option("-d","--debug",action="callback",callback=d_action,dest="debug",default=False,help="""\
enable debugging message output (implies --verbose).""")



def l_action(option,opt_str,value,parser):
    """Callback function for the -l option."""
    boolean_option_action(option,opt_str,value,parser)
    from topo.misc.legacy import install_legacy_support
    print "Enabling legacy support."
    install_legacy_support()
    
topo_parser.add_option("-l","--legacy",action="callback",callback=l_action,dest="legacy",default=False,help="""\
launch Topographica with legacy support enabled.""")


def gui(start=True):
    """Start the GUI as if -g were supplied in the command used to launch Topographica."""
    if matplotlib_imported: 
        rcParams['backend']='TkAgg'
    auto_import_commands()
    if start:
        import topo.tkgui
        topo.tkgui.start()


# Topographica stays open if an error occurs after -g
# (see comment by i_action)
def g_action(option,opt_str,value,parser):
    """Callback function for the -g option."""
    boolean_option_action(option,opt_str,value,parser)
    interactive()
    gui()


topo_parser.add_option("-g","--gui",action="callback",callback=g_action,dest="gui",default=False,help="""\
launch an interactive graphical user interface; \
equivalent to -c 'from topo.misc.commandline import gui ; gui()'. \
Implies -a.""")

topo_parser.add_option("--pdb",action="store_true",dest="pdb",help="""\
Automatically call the pdb debugger after every uncaught \
exception. See IPython documentation for further details.""")

# Keeps track of whether something has been performed, when deciding whether to assume -i
something_executed=False

def c_action(option,opt_str,value,parser):
    """Callback function for the -c option."""
    #print "Processing %s '%s'" % (opt_str,value)
    exec value in __main__.__dict__
    global something_executed
    something_executed=True
    openmp_settings_names = ['openmp_threads', 'openmp_min_threads', 'openmp_max_threads']
    openmp_present = [(k in __main__.__dict__) for k in openmp_settings_names]
    if openmp_present and parser.values.gui: 
        print "\nWARNING: For OpenMP settings to be used properly they need to be specified after the -g flag."
    
topo_parser.add_option("-c","--command",action = "callback",callback=c_action,type="string",
		       default=[],dest="commands",metavar="\"<command>\"",
		       help="string of arbitrary Python code to be executed in the main namespace.")



def p_action(option,opt_str,value,parser):
    """Callback function for the -p option."""
    global_params.exec_in_context(value)
    global something_executed
    something_executed=True
            
topo_parser.add_option("-p","--set-parameter",action = "callback",callback=p_action,type="string",
		       default=[],dest="commands",metavar="\"<command>\"",
		       help="command specifying value(s) of script-level (global) Parameter(s).")


def auto_import_commands():
    """Import the contents of all files in the topo/command/ directory."""
    import re,os
    import topo
    import __main__

    # CEBALERT: this kind of thing (topo.__file__) won't work with
    # py2exe and similar tools
    topo_path = os.path.join(os.path.split(topo.__file__)[0],"command")
    for f in os.listdir(topo_path):
        if re.match('^[^_.].*\.py$',f):
            modulename = re.sub('\.py$','',f)
            exec "from topo.command."+modulename+" import *" in __main__.__dict__
    exec "from topo.command import *" in __main__.__dict__
    
def a_action(option,opt_str,value,parser):
    """Callback function for the -a option."""
    auto_import_commands()
    
topo_parser.add_option("-a","--auto-import-commands",action="callback",callback=a_action,help="""\
import everything from commands/*.py into the main namespace, for convenience; \
equivalent to -c 'from topo.misc.commandline import auto_import_commands ; auto_import_commands()'.""")



def exec_startup_files():
    """
    Execute startup files.

    Linux/UNIX/OS X: ~/.topographicarc
    Windows: %USERPROFILE%\topographica.ini
    """
    # From Bilal: On OS X, ~/Library/Preferences/ is the standard path
    # for user-defined params. The filename format (corresponding to
    # .ini on windows) is org.topographica.plist, where a plist is an
    # XML file. But, many shell-based programs follow the Unix
    # convention, so we should be fine doing that.

    # Linux/UNIX/OS X:
    rcpath = os.path.join(os.path.expanduser("~"),'.topographicarc')
    # Windows (ini is convention, and can be double clicked to edit):
    inipath = os.path.join(os.path.expandvars("$USERPROFILE"),'topographica.ini')

    for startup_file in (rcpath,inipath):
        if os.path.exists(startup_file):
            print "Executing user startup file %s" % (startup_file)
            execfile(startup_file,__main__.__dict__)
    
    #####
    # CEBALERT: locations we used to use on Windows and OS X. Should
    # remove after 0.9.8.
    # application data on windows
    inipath = os.path.join(os.path.expandvars("$APPDATA"),'Topographica','topographica.ini')
    # application support on OS X  
    configpath = os.path.join(os.path.expanduser("~"),"Library","Application Support",'Topographica','topographica.config')
    for startup_file in (configpath,inipath):
        if os.path.exists(startup_file):
            param.Parameterized().warning("Ignoring %s; location for startup file is %s (UNIX/Linux/Mac OS X) or %s (Windows)."%(startup_file,rcpath,inipath)) 
    #####


def get_omp_num_threads(openmp_threads, openmp_min_threads, openmp_max_threads):
    """ Helper function to implement sensible OpenMP behaviour where
        possible.  Returns a integer tuple (OpenMP threads, CPUs detected).

        CPUs detected may be None if multiprocessing is not available (ie
        Python 2.5) and consequently OpenMP threads may be None if the user
        gave a relative thread specification (ie openmp_threads <= 0). When
        OpenMP threads is None, the default OpenMP behaviour of using all
        available cores is used.
        """

    if 'OMP_NUM_THREADS' in os.environ: return (None, None)

    if 'NSLOTS' in os.environ: return ('NSLOTS', None)

    try:
        import multiprocessing
        total_cores = multiprocessing.cpu_count()
        if total_cores == 1: return (1, 1)
        if total_cores == 2: return (2, 2)
    except:
        print "Cannot import multiprocessing to determine number of cores."
        total_cores = None

    if total_cores and (-openmp_threads >= (total_cores-1)):
        print "OpenMP: Topographica needs a positive number of threads to execute!"
        return (1, total_cores)

    if (openmp_threads <= 0) and total_cores:      openmp_threads = total_cores + openmp_threads
    if (openmp_threads <= 0) and not total_cores:  return (None, total_cores)

    if (total_cores is not None) and (openmp_threads > total_cores):
        print "OpenMP: More threads specified than cores detected."; return (total_cores, total_cores)

    if openmp_threads < openmp_min_threads:
        print "OpenMP: Using minimum number of allowed threads."
        openmp_threads = openmp_min_threads

    if (openmp_max_threads is None): return (openmp_threads, total_cores)

    elif (openmp_max_threads < openmp_min_threads):
        print"OpenMP: Maximum allowed threads lower than minimum allowed threads. Ignoring maximum limit."
        return (openmp_threads, total_cores)
    elif (openmp_threads > openmp_max_threads):
        print "OpenMP: Using maximum number of allowed threads."
        return (openmp_max_threads, total_cores)
    else:
        return (openmp_threads, total_cores)


### Execute what is specified by the options.

def process_argv(argv):
    """
    Process command-line arguments (minus argv[0]!), rearrange and execute.
    """
    # Initial preparation
    import __main__
    for (k,v) in global_constants.items():
        exec '%s = %s' % (k,v) in __main__.__dict__

    exec_startup_files()

    # Repeatedly process options, if any, followed by filenames, if any, until nothing is left
    topo_parser.disable_interspersed_args()
    args=argv
    option=None
    global something_executed
    while True:
        # Process options up until the first filename
        (option,args) = topo_parser.parse_args(args,option)

        # Handle filename
        if args:
            filename=args.pop(0)
            #print "Executing %s" % (filename)
            filedir = os.path.dirname(os.path.abspath(filename))
            sys.path.insert(0,filedir) # Allow imports relative to this file's path
            sim_name_from_filename(filename) # Default value of topo.sim.name

            execfile(filename,__main__.__dict__)
            something_executed=True

        if not args:
            break

    global_params.check_for_unused_names()

    # OpenMP settings and defaults
    openmp_threads = __main__.__dict__.get('openmp_threads')
    if (openmp_threads is None): openmp_threads=-1

    openmp_min_threads = __main__.__dict__.get('openmp_min_threads')
    if (openmp_min_threads is None): openmp_min_threads=2

    openmp_max_threads = __main__.__dict__.get('openmp_max_threads')

    if (openmp_threads != 1): # OpenMP is disabled if openmp_threads == 1

        (num_threads, total_cores) = get_omp_num_threads(openmp_threads,
                                                         openmp_min_threads,
                                                         openmp_max_threads)

        if num_threads is None:
            print "OpenMP: Using OMP_NUM_THREADS environment variable if set. Otherwise, all cores in use."
        elif num_threads == 'NSLOTS':
            os.environ['OMP_NUM_THREADS'] =  os.environ['NSLOTS']
            print "NSLOTS environment variable found; overriding any other thread settings and using N=%s threads" % os.environ['NSLOTS']

        elif total_cores is None:
            print "OpenMP: Using %d threads" % num_threads
            os.environ['OMP_NUM_THREADS'] =  str(num_threads)
        else:
            print "OpenMP: Using %d threads on a machine with %d detected CPUs" % (num_threads, total_cores)
            os.environ['OMP_NUM_THREADS'] =  str(num_threads)

    # If no scripts and no commands were given, pretend -i was given.
    if not something_executed: interactive()

    if option.gui: topo.guimain.title(topo.sim.name)

    ## INTERACTIVE SESSION BEGINS HERE (i.e. can't have anything but
    ## some kind of cleanup code afterwards)
    if os.environ.get('PYTHONINSPECT'):
        print "Output path: %s" % param.normalize_path.prefix
        print BANNER    
        # CBALERT: should probably allow a way for users to pass
        # things to IPython? Or at least set up some kind of
        # topographica ipython config file. Right now, a topo_parser
        # option has to be added for every ipython option we want to
        # support (e.g. see --pdb)

        if ipython_shell_interface == "IPython.Shell":
            # IPython 0.10 and earlier

            # Stop IPython namespace hack?
            # http://www.nabble.com/__main__-vs-__main__-td14606612.html
            __main__.__name__="__mynamespace__"

            ipython_args = ['-noconfirm_exit','-nobanner',
                            '-pi1',CommandPrompt.get_format(),
                            '-pi2',CommandPrompt2.get_format(),
                            '-po',OutputPrompt.get_format()]
            if option.pdb:
                ipython_args.append('-pdb')

            ipshell = IPShell(ipython_args,user_ns=__main__.__dict__)
            ipshell.mainloop(sys_exit=1)

        elif ipython_shell_interface == "InteractiveShellEmbed":
            # IPython 0.11 and later

            config = Config()

            if ipython_prompt_interface == "PromptManager":
                config.PromptManager.in_template = CommandPrompt.get_format()
                config.PromptManager.in2_template = CommandPrompt2.get_format()
                config.PromptManager.out_template = OutputPrompt.get_format()
            else:
                config.InteractiveShell.prompt_in1 = CommandPrompt.get_format()
                config.InteractiveShell.prompt_in2 = CommandPrompt2.get_format()
                config.InteractiveShell.prompt_out = OutputPrompt.get_format()
            config.InteractiveShell.confirm_exit = False
            ipshell = IPShell(config=config,user_ns=__main__.__dict__,
                              banner1="",exit_msg="")
            if option.pdb:
                ipshell.call_pdb = True
            ipshell()
            sys.exit()



"""
Unit test for the matplot lib in tkgui
$Id$
"""

import unittest
import matplotlib

class TestMatPlotLibTk(unittest.TestCase):

    def test_TkAgg(self):
        import matplotlib
        #matplotlib.use('TkAgg')
        try: import pylab
        except ImportError: havetkagg = False
        else: havetkagg = True
        if not havetkagg:
            raise 'TkAgg is not available.  When MatPlotLib was built it did not find Tkinter.'



suite = unittest.TestSuite()
#  Uncomment the following line of code, to disable the test if
#  $DISPLAY is undefined.  Used mainly for GUI testing.
suite.requires_display = True
suite.addTest(unittest.makeSuite(TestMatPlotLibTk))



"""
A family of high-level user commands acting on the entire simulation.

Any new commands added to this directory will automatically become
available for any program.

Commands here should be 'bullet-proof' and work 'from scratch'.
That is, they should print warnings if required but should not raise
errors that would interrupt e.g. a long batch run of simulation work,
no matter what the context from which they are called.

$Id$
"""

import cPickle as pickle

import os,sys,re,string,time,platform

import __main__

# gzip module might not have been built (if zlib could not be found when building)
try:
    import gzip
except ImportError:
    pass

import param
from param.parameterized import ParameterizedFunction, ParamOverrides
from param import normalize_path

import topo
from topo.base.sheet import Sheet
from topo.base.projection import ProjectionSheet
from topo.sheet import GeneratorSheet
from topo.misc.util import MultiFile
from topo.misc.picklemain import PickleMain
from topo.misc.snapshots import PicklableClassAttributes
from topo.misc.genexamples import generate as _generate
from topo.base.functionfamily import PatternDrivenAnalysis

# CEBALERT: Could say min Python 2.6 and use Python's own. Or is it
# 2.7?
from topo.misc.odict import OrderedDict


def generate_example(target):
    """
    Generate the saved network target, as defined in
    topo.misc.genexamples.
    """
    _generate(targets=[target])


# Not sure where to put CommandMetaclass, since it doesn't need to be
# seen or used by anyone. Probably also the import error classes
# shouldn't be so visible.
from param.parameterized import ParameterizedMetaclass
class CommandMetaclass(ParameterizedMetaclass):
    """
    A class having this as a metaclass will have its __call__() method
    automatically wrapped so that any exception occurring inside
    __call__() will be passed to the class's _except() method.
    """
    def __new__(mcs,classname,bases,classdict):        
        if '__call__' in classdict:
            classdict['__call__'] = mcs._safecall(classdict['__call__'])
            #assert '_except' in classdic
        # else it is probably abstract, or something
        return ParameterizedMetaclass.__new__(mcs,classname,bases,classdict)

    @classmethod
    def _safecall(mcs,fn):
        """
        Wrap fn with caller, which catches any exception raised inside
        fn() and passes it to _except().
        """
        def caller(self,*args,**kw):
            try:
                return fn(self,*args,**kw)
            except Exception, e:
                # Mis-invoked call should raise error as normal.
                if isinstance(e,TypeError):
                    import inspect
                    # Is this a hack to detect mis-calling?
                    if len(inspect.getargspec(fn)[0])!=len(args):
                        raise
                try:
                    return self._except(e)
                except:
                    # the _except() method raises an error (or doesn't
                    # exist). what should happen? programming error,
                    # so I guess just re-raise
                    raise
        return caller
                


class Command(ParameterizedFunction):
    """
    Parameterized command: any error when the command is run (called)
    will not raise an exception, but will instead generate a warning.
    """
    __metaclass__ = CommandMetaclass
    __abstract = True

    def _except(self,e):
        # import traceback
        # print traceback.print_exc()
        self.warning("%s failed: %s"%(self,e)) # or traceback.format_exc()))
        
    def __call__(self,*args,**params):
        return super(Command,self).__call__(*args,**params)


class ImportErrorObject(object):
    """
    Raises an ImportError on any attempt to access an attribute, call,
    or get an item.

    Useful to delay an ImportError until the point of use, thus
    allowing e.g. a class attribute to contain something from a
    non-core external module (e.g. pylab). 

    Delaying an ImportError until the point of use allows users to be
    informed of the possibility of having various extra functions on
    installation of a missing package.
    """
    def __init__(self,module_name):
        self.__module_name = module_name
    def _raise(self):
        #param.Parameterized().warning("err:%s"%self.module_name)
        raise ImportError, "No module named %s. Install %s to get this functionality."%(self.__module_name,self.__module_name)
        return None
    def __call__(self,*args,**kw):
        self._raise()
    # Might be better to override __getattribute__, special casing the
    # module_name attribute. Then everything is guaranteed to raise an
    # error (rather than covering call, getitem, getattr, and maybe
    # other things I've forgotten about).
    def __getattr__(self,name):
        return self._raise()
    def __getitem__(self,i):
        self._raise()



class ImportErrorRaisingFakeModule(object):
    """
    Returns an ImportErrorObject for any attribute request.

    Instances of this class can be used in place of a module to delay
    an import error until the point of use of an attribute of that
    module.

    See ImportErrorObject for more details.
    """
    def __init__(self,module_name):
        self.__module_name = module_name
    def __getattr__(self,name):
        return ImportErrorObject(self.__module_name)


# CEBALERT: commands in here should inherit from Command, and make use
# of _except() to ensure all necessary state is reverted.

def save_input_generators():
    """Save a copy of the active_sim's current input_generators for all GeneratorSheets."""
    # ensure EPs get started (if save_input_generators is called before the simulation is run())
    topo.sim.run(0.0) 

    generator_sheets = topo.sim.objects(GeneratorSheet).values()
    for sheet in generator_sheets:
        sheet.push_input_generator()


def restore_input_generators():
    """Restore previously saved input_generators for all of topo.sim's GeneratorSheets."""
    generator_sheets = topo.sim.objects(GeneratorSheet).values()
    for sheet in generator_sheets:
        sheet.pop_input_generator()


def clear_event_queue():
    """Remove pending events from the simulator's event queue."""
    topo.sim.event_clear()


def pattern_present(inputs={},duration=1.0,plastic=False,overwrite_previous=False,apply_output_fns=True):
    """
    Present the specified test patterns for the specified duration.

    Given a set of input patterns (dictionary of
    GeneratorSheetName:PatternGenerator pairs), installs them into the
    specified GeneratorSheets, runs the simulation for the specified
    length of time, then restores the original patterns and the
    original simulation time.  Thus this input is not considered part
    of the regular simulation, and is usually for testing purposes.

    As a special case, if 'inputs' is just a single pattern, and not
    a dictionary, it is presented to all GeneratorSheets.
    
    If a simulation is not provided, the active simulation, if one
    exists, is requested.

    If this process is interrupted by the user, the temporary patterns
    may still be installed on the retina.

    If overwrite_previous is true, the given inputs overwrite those
    previously defined.

    If plastic is False, overwrites the existing values of Sheet.plastic
    to disable plasticity, then reenables plasticity.
    
    In order to to see the sequence of values presented, use the back arrow 
    history mechanism in the GUI. Note that the GUI's Activity window must 
    be open and the display parameter set to true (display=True).
    """
    # ensure EPs get started (if pattern_present is called before the simulation is run())
    topo.sim.run(0.0) 
    
       
    if not overwrite_previous:
        save_input_generators()

    if not plastic:
        # turn off plasticity everywhere
        for sheet in topo.sim.objects(Sheet).values():
             sheet.override_plasticity_state(new_plasticity_state=False)

    if not apply_output_fns:
        for each in topo.sim.objects(Sheet).values():
            if hasattr(each,'measure_maps'):
               if each.measure_maps: 
                   each.apply_output_fns = False

    # Register the inputs on each input sheet
    generatorsheets = topo.sim.objects(GeneratorSheet)
    if not isinstance(inputs,dict):
        for g in generatorsheets.values():
            g.set_input_generator(inputs)
    else:
        for each in inputs.keys():
            if generatorsheets.has_key(each):
                generatorsheets[each].set_input_generator(inputs[each])
            else:
                param.Parameterized().warning(
                    '%s not a valid Sheet name for pattern_present.' % each)

    topo.sim.event_push()
    # CBENHANCEMENT: would be nice to break this up for visualizing motion
    topo.sim.run(duration) 
    topo.sim.event_pop()

    # turn sheets' plasticity and output_fn plasticity back on if we turned it off before

    if not plastic:
        for sheet in topo.sim.objects(Sheet).values():
            sheet.restore_plasticity_state()
          
    if not apply_output_fns:
        for each in topo.sim.objects(Sheet).values():
            each.apply_output_fns = True
 
        
    if not overwrite_previous:
        restore_input_generators()


# This class is left around to support older snapshots: All snapshots
# since 0.9.7 up until r11545 (addition of UnpickleEnvironmentCreator)
# have a pickled instance of this class. We maintain the same behavior
# as before for them: install all legacy support.
class _VersionPrinter(object):
    def __setstate__(self,state):
        import topo.misc.legacy as L
        L.SnapshotSupport.install("0.9.7")


class UnpickleEnvironmentCreator(object):
    """When unpickled, installs any necessary legacy support."""
    def __init__(self,release,version):
        self.release = release
        self.version = version
    def __getstate__(self):
        return {'release':self.release,
                'version':self.version}
    def __setstate__(self,state):
        self.release = state['release']
        self.version = state['version']
        import topo.misc.legacy as L
        L.SnapshotSupport.install(self.release,self.version)



def save_snapshot(snapshot_name=None):
    """
    Save a snapshot of the network's current state.

    The snapshot is saved as a gzip-compressed Python binary pickle.

    As this function uses Python's 'pickle' module, it is subject to
    the same limitations (see the pickle module's documentation) -
    with the notable exception of class attributes. Python does not
    pickle class attributes, but this function stores class attributes
    of any Parameterized class that is declared within the topo
    package. See the param.parameterized.PicklableClassAttributes
    class for more information.
    """
    if not snapshot_name:
        snapshot_name = topo.sim.basename() + ".typ"

    # For now we just search topo, but could do same for other packages.

    # CEBALERT: shouldn't it be topo and param? I guess we already get
    # many classes defined in param because they are imported into
    # topo at some point anyway.
    topoPOclassattrs = PicklableClassAttributes(topo,exclusions=('plotting','tests','tkgui'),
                                                startup_commands=topo.sim.startup_commands)

    from topo.misc.commandline import global_params

    topo.sim.RELEASE=topo.release
    topo.sim.VERSION=topo.version

    to_save = (UnpickleEnvironmentCreator(topo.release,topo.version),
               PickleMain(),
               global_params,
               topoPOclassattrs,
               topo.sim)

    try:
        snapshot_file=gzip.open(normalize_path(snapshot_name),'wb',compresslevel=5)
    except NameError:
        snapshot_file=open(normalize_path(snapshot_name),'wb')

    pickle.dump(to_save,snapshot_file,2)

    
    snapshot_file.close()


def load_snapshot(snapshot_name):
    """
    Load the simulation stored in snapshot_name.
    """
    # unpickling the PicklableClassAttributes() executes startup_commands and
    # sets PO class parameters.

    snapshot_name = param.resolve_path(snapshot_name)

    # If it's not gzipped, open as a normal file.
    try:
        snapshot = gzip.open(snapshot_name,'r')
        snapshot.read(1)
        snapshot.seek(0)
    except (IOError,NameError):
        snapshot = open(snapshot_name,'r')
    
    try:
        pickle.load(snapshot)
    except ImportError:
        # CEBALERT: Support snapshots where the unpickling support
        # (UnpickleEnvironmentCreator) cannot be found because the
        # support itself was moved from topo.command.basic to
        # topo.command.__init__! Was it a bad idea to have the support
        # code loaded via an object?
        sys.modules['topo.command.basic'] = topo.command
        # Could instead set find_global on cPickle.Unpickler (could
        # support all import changes that way, as alternative to what
        # we currently do), but I'm reluctant to mess with cPickle's
        # default way of finding things. (Also it would be specific to
        # cPickle; would be different for pickle.)
        
        snapshot.seek(0)
        try:
            pickle.load(snapshot)
        except:
            import traceback

            m = """
            Snapshot could not be loaded.

            Please file a support request via topographica.org.

Loading error:
%s
            """%traceback.format_exc()

            param.Parameterized(name="load_snapshot").warning(m)


    snapshot.close()

    # Restore subplotting prefs without worrying if there is a
    # problem (e.g. if topo/analysis/ is not present)
    try: 
        from topo.analysis.featureresponses import Subplotting
        Subplotting.restore_subplots()
    except:
        p = param.Parameterized(name="load_snapshot")
        p.message("Unable to restore Subplotting settings")




def save_script_repr(script_name=None):
    """
    Save the current simulation as a Topographica script.

    Generates a script that, if run, would generate a simulation with
    the same architecture as the one currently in memory.  This can be
    useful when defining networks in place, so that the same general
    configuration can be recreated later.  It also helps when
    comparing two similar networks generated with different scripts,
    so that the corresponding items can be matched rigorously.

    Note that the result of this operation is usually just a starting
    point for further editing, because it will not usually be runnable
    as-is (for instance, some parameters may not have runnable
    representations).  Even so, this is usually a good start.
    """
    if not script_name:
        script_name = topo.sim.basename() + "_script_repr.ty"
        
    header = ("# Generated by Topographica %s on %s\n\n" %
              (topo.release,time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime())))
    script = header+topo.sim.script_repr()
    
    script_file = open(normalize_path(script_name),'w')
    script_file.write(script)


# Location of the version-controlled topographica directory (i.e. path
# of topo/ but up a level). Could be None. Nothing should assume that
# there is a version control system available.
try:
    vc_topographica_dir = os.path.split(os.path.split(topo.__file__)[0])[0]
except:
    vc_topographica_dir = None


# decorator that changes to vc_topographica_dir for duration of fn,
# if there is such a directory. Otherwise, doesn't change directory.
def in_vc_topographica_dir(fn):
    import os
    def temporarily_change_to_vc_topographica_dir(*args,**kw):
        orig_path = os.getcwd()
        if vc_topographica_dir is not None:
            os.chdir(vc_topographica_dir)
        try:
            result = fn(*args,**kw)
        finally:
            # ensure dir put back even if there's an error calling fn
            if os.getcwd()!=orig_path:
                os.chdir(orig_path)
        return result
    return temporarily_change_to_vc_topographica_dir


@in_vc_topographica_dir
def _get_vc_commands():
    # return name of version control system (None if no vc could be
    # detected)
    import os.path
    vc_types = {'git':["status","diff",["log","-n1"],["svn","log","--limit=1"]],
                'svn':["info","status","diff"],
                'bzr':['info','status','diff']}
    for vc_type,commands in vc_types.items():
        if os.path.exists(".%s"%vc_type):
            return vc_type,commands

@in_vc_topographica_dir
def _print_vc_info(filename):
    """Save the version control status of the current code to the specified file."""
    try:
        import subprocess
        f = open(normalize_path(filename),'w')
        f.write("Information about working copy used for batch run\n\n")
        f.write("topo.version=%s\n"% topo.version)
        f.flush()
        vctype,commands = _get_vc_commands()
        for cmd in commands:
            fullcmd = [vctype,cmd] if isinstance(cmd,str) else [vctype]+cmd
        
            # Note that we do not wait for the process below to finish
            # (by calling e.g. wait() on the Popen object). Although
            # this was probably done unintentionally, for a slow svn
            # connection, it's an advantage. But it does mean the
            # output of each command can appear in the file at any
            # time (i.e. the command outputs appear in the order of
            # finishing, rather than in the order of starting, making
            # it impossible to label the commands).
            subprocess.Popen(fullcmd,stdout=f,stderr=subprocess.STDOUT)
    except:
        print "Unable to retrieve version control information."
    finally:
        f.close()


def _save_parameters(p,filename):
    from topo.misc.commandline import global_params
    
    g = {'global_params_specified':p,
         'global_params_all':dict(global_params.get_param_values())}

    for d in g.values():
        if 'name' in d:
            del d['name']
        if 'print_level' in d:
            del d['print_level']

    pickle.dump(g,open(normalize_path(filename),'w'))
     
# I'd expect your personal name_replacements to be set in some file
# you use to create batch runs, but it can alsp be set on the
# commandline. Before calling run_batch(), include something like the
# following:
# run_batch.dirname_params_filter.map=OrderedDict(("cortex_density","cd"))

class param_formatter(ParameterizedFunction):

    # CEBALERT: should I have made this a parameter at the run_batch
    # level? And I don't know what to call it.
    map = param.Dict(default=OrderedDict(),doc="""
        Optional ordered dictionary of alternative names to use for
        parameters, parameter_name:alternative_name

        Use to shorten common parameter names (directory names are
        limited in length on most file systems), and to specify an
        order.

        Names not specified here will be sorted alphabetically.""")

    def __call__(self,params):
        result = ""
        # present in params but not in map
        unspecified_in_map = sorted(set(params).difference(set(self.map)))
        # present in params and in map, preserving order of map
        specified_in_map = [n for n in self.map.keys() if n in params]

        for pname in specified_in_map+unspecified_in_map:
            val = params[pname]
            # Special case to give reasonable filenames for lists
            valstr= ("_".join([str(i) for i in val]) if isinstance(val,list)
                     else str(val))
            result += "," + self.map.get(pname,pname) + "=" + valstr
        return result


# Used only by default_analysis_function
# Should be in order they are needed; e.g. Activity after map measurement,
# in case Activity plot includes map subplots
default_analysis_plotgroups=["Orientation Preference","Activity"]

def default_analysis_function():
    """
    Basic example of an analysis command for run_batch; users are
    likely to need something similar but highly customized.
    """
    # CEBALERT: why are these imports here rather than at the top?
    import topo
    from topo.command.analysis import save_plotgroup

    # Save all plotgroups listed in default_analysis_plotgroups
    for pg in default_analysis_plotgroups:
        save_plotgroup(pg,use_cached_results=True)

    # Plot projections from each measured map
    measured_sheets = [s for s in topo.sim.objects(ProjectionSheet).values()
                       if hasattr(s,'measure_maps') and s.measure_maps]
    for s in measured_sheets:
        for p in s.in_connections:
            save_plotgroup("Projection",projection=p)

    # Test response to a standardized pattern
    from topo.pattern import Gaussian
    from math import pi
    pattern_present(inputs=Gaussian(orientation=pi/4,aspect_ratio=4.7))
    save_plotgroup("Activity",saver_params={"filename_suffix":"_45d"})


def load_kwargs(fname, glob, loc, fail_exception=False):
    """ 
    Helper function to allow keyword arguments (dictionary format)
    to be  loaded from a file 'fname'. The intended use is to allow a callable
    (specifically run_batch) to obtain its settings and parameters from file. 

    This is useful when dispatching jobs on a cluster as you can then queue
    run_batch jobs (eg. using qsub) before all the settings are known. This
    type of scenario is typical in parameter search (eg hillclimbing) where
    the settings file for future run_batch instances are conditional on data
    from previous simulations.

    Variable glob should be provided as globals() and loc should be provided
    as locals(). Either a dictionary is returned or an exception is raised
    (conditioned on fail_exception). If fail_exception=False and eval does 
    not evaluateas expected, an empty dictionary is returned. Eval is used
    as it allows classes, objects and other complex datastructures to load.
    """
    with open(fname,'r') as f: lines = f.readlines()
    expression = "".join([l.strip() for l in lines])
    kwargs = eval(expression, glob, loc)
    if not isinstance(kwargs,dict): 
        if fail_exception: raise Exception('Invalid settings file.')
        else:              return {}
    else:                        return kwargs


# ALERT: Need to move docs into params.
class run_batch(ParameterizedFunction):
    """
    Run a Topographica simulation in batch mode.

    Features:

      - Generates a unique, well-defined name for each 'experiment'
        (i.e. simulation run) based on the date, script file, and
        parameter settings. Note that very long names may be truncated
        (see the max_name_length parameter).

      - Allows parameters to be varied on the command-line,
        to allow comparing various settings

      - Saves a script capturing the simulation state periodically,
        to preserve parameter values from old experiments and to allow
        them to be reproduced exactly later

      - Can perform user-specified analysis routines periodically,
        to monitor the simulation as it progresses.

      - Stores commandline output (stdout) in the output directory
        
    A typical use of this function is for remote execution of a large
    number of simulations with different parameters, often on remote
    machines (such as clusters).
    
    The script_file parameter defines the .ty script we want to run in
    batch mode. The output_directory defines the root directory in
    which a unique individual directory will be created for this
    particular run.  The optional analysis_fn can be any python
    function to be called at each of the simulation iterations defined
    in the analysis times list.  The analysis_fn should perform
    whatever analysis of the simulation you want to perform, such as
    plotting or calculating some statistics.  The analysis_fn should
    avoid using any GUI functions (i.e., should not import anything
    from topo.tkgui), and it should save all of its results into
    files.

    As a special case, a number can be passed for the times list, in
    which case it is used to scale a default list of times up to
    10000; e.g. times=2 will select a default list of times up to
    20000.  Alternatively, an explicit list of times can be supplied.

    Any other optional parameters supplied will be set in the main
    namespace before any scripts are run.  They will also be used to
    construct a unique topo.sim.name for the file, and they will be
    encoded into the simulation directory name, to make it clear how
    each simulation differs from the others.

    If requested by setting snapshot=True, saves a snapshot at the
    end of the simulation.

    If available and requested by setting vc_info=True, prints
    the revision number and any outstanding diffs from the version
    control system.

    Note that this function alters param.normalize_path.prefix so that
    all output goes into the same location. The original value of
    param.normalize_path.prefix is deliberately not restored at the
    end of the function so that the output of any subsequent commands
    will go into the same place.
    """
    output_directory=param.String("Output")

    analysis_fn = param.Callable(default_analysis_function)
    
    times = param.Parameter(1.0)

    snapshot=param.Boolean(True)

    vc_info=param.Boolean(True)

    dirname_prefix = param.String(default="",doc="""
        Optional prefix for the directory name (allowing e.g. easy
        grouping).""")

    tag = param.String(default="",doc="""
        Optional tag to embed in directory prefix to allow unique
        directory naming across multiple independent batches that
        share a common timestamp.""")

    # CB: do any platforms also have a maximum total path length?
    max_name_length = param.Number(default=200,doc="""
        The experiment's directory name will be truncated at this
        number of characters (since most filesystems have a
        limit).""")

    name_time_format = param.String(default="%Y%m%d%H%M",doc="""
        String format for the time included in the output directory
        and file names.  See the Python time module library
        documentation for codes.
        
        E.g. Adding '%S' to the default would include seconds.""")

    timestamp = param.NumericTuple(default=(0,0), doc="""
        Optional override of timestamp in Python struct_time 8-tuple format. 
        Useful when running many run_batch commands as part of a group with
        a shared timestamp. By default, the timestamp used is the time when
        run_batch is started.""") 

    save_global_params = param.Boolean(default=True,doc="""
        Whether to save the script's global_parameters to a pickle in
        the output_directory after the script has been loaded (for
        e.g. future inspection of the experiment).""")

    dirname_params_filter = param.Callable(param_formatter.instance(),doc="""
        Function to control how the parameter names will appear in the
        output_directory's name.""")


    def _truncate(self,p,s):
        """
        If s is greater than the max_name_length parameter, truncate it
        (and indicate that it has been truncated).
        """
        # '___' at the end is supposed to represent '...'
        return s if len(s)<=p.max_name_length else s[0:p.max_name_length-3]+'___' 
                
    def __call__(self,script_file,**params_to_override):
        p=ParamOverrides(self,params_to_override,allow_extra_keywords=True)

        import os
        import shutil
    
        # Construct simulation name, etc.
        scriptbase= re.sub('.ty$','',os.path.basename(script_file))
        prefix = ""
        if p.timestamp==(0,0): prefix += time.strftime(p.name_time_format)
        else:                  prefix += time.strftime(p.name_time_format, p.timestamp)

        prefix += "_" + scriptbase + "_" + p.tag
        simname = prefix

        # Construct parameter-value portion of filename; should do more filtering
        # CBENHANCEMENT: should provide chance for user to specify a
        # function (i.e. make this a function, and have a parameter to
        # allow the function to be overridden).
        # And sort by name by default? Skip ones that aren't different
        # from default, or at least put them at the end?
        prefix += p.dirname_params_filter(p.extra_keywords())

        # Set provided parameter values in main namespace
        from topo.misc.commandline import global_params
        global_params.set_in_context(**p.extra_keywords())
    
        # Create output directories
        if not os.path.isdir(normalize_path(p['output_directory'])):
            try: os.mkdir(normalize_path(p['output_directory'])) 
            except OSError: pass   # Catches potential race condition (simultaneous run_batch runs)
        
        dirname = self._truncate(p,p.dirname_prefix+prefix)
        normalize_path.prefix = normalize_path(os.path.join(p['output_directory'],dirname))
        
        if os.path.isdir(normalize_path.prefix):
            print "Batch run: Warning -- directory already exists!"
            print "Run aborted; wait one minute before trying again, or else rename existing directory: \n" + \
                  normalize_path.prefix
    
            sys.exit(-1)
        else:
            os.mkdir(normalize_path.prefix)
            print "Batch run output will be in " + normalize_path.prefix
    
    
        if p['vc_info']:
            _print_vc_info(simname+".diffs")
    
        hostinfo = "Host: " + " ".join(platform.uname())
        topographicalocation = "Topographica: " + os.path.abspath(sys.argv[0])
        topolocation = "topo package: " + os.path.abspath(topo.__file__)
        scriptlocation = "script: " + os.path.abspath(script_file)

        starttime=time.time()
        startnote = "Batch run started at %s." % time.strftime("%a %d %b %Y %H:%M:%S +0000",
                                                               time.gmtime())

        # store a re-runnable copy of the command used to start this batch run
        try:
            # pipes.quote is undocumented, so I'm not sure which
            # versions of python include it (I checked python 2.6 and
            # 2.7 on linux; they both have it).
            import pipes
            quotefn = pipes.quote            
        except (ImportError,AttributeError):
            # command will need a human to insert quotes before it can be re-used
            quotefn = lambda x: x
            
        command_used_to_start = string.join([quotefn(arg) for arg in sys.argv])

        # CBENHANCEMENT: would be nice to separately write out a
        # runnable script that does everything necessary to
        # re-generate results (applies diffs etc).

        # Shadow stdout to a .out file in the output directory, so that
        # print statements will go to both the file and to stdout.
        batch_output = open(normalize_path(simname+".out"),'w')
        batch_output.write(command_used_to_start+"\n")
        sys.stdout = MultiFile(batch_output,sys.stdout)
    
        print
        print hostinfo
        print topographicalocation
        print topolocation
        print scriptlocation
        print
        print startnote
    
        from topo.misc.commandline import auto_import_commands
        auto_import_commands()
        
        # Ensure that saved state includes all parameter values
        from topo.command import save_script_repr
        param.parameterized.script_repr_suppress_defaults=False
    
        # Save a copy of the script file for reference
        shutil.copy2(script_file, normalize_path.prefix)
        shutil.move(normalize_path(scriptbase+".ty"),
                    normalize_path(simname+".ty"))
    
        
        # Default case: times is just a number that scales a standard list of times
        times=p['times']
        if not isinstance(times,list):
            times=[t*times for t in [0,50,100,500,1000,2000,3000,4000,5000,10000]]
    
        # Run script in main
        error_count = 0
        initial_warning_count = param.parameterized.warning_count
        try:
            execfile(script_file,__main__.__dict__) #global_params.context
            global_params.check_for_unused_names()
            if p.save_global_params:
                _save_parameters(p.extra_keywords(),simname+".global_params.pickle")
            print_sizes()
            topo.sim.name=simname
    
            # Run each segment, doing the analysis and saving the script state each time
            for run_to in times:
                topo.sim.run(run_to - topo.sim.time())
                p['analysis_fn']()
                save_script_repr()
                elapsedtime=time.time()-starttime
                param.Parameterized(name="run_batch").message(
                    "Elapsed real time %02d:%02d." % (int(elapsedtime/60),int(elapsedtime%60)))
    
            if p['snapshot']:
               save_snapshot()
                
        except:
            error_count+=1
            import traceback
            traceback.print_exc(file=sys.stdout)
            sys.stderr.write("Warning -- Error detected: execution halted.\n")
    
    
        print "\nBatch run completed at %s." % time.strftime("%a %d %b %Y %H:%M:%S +0000",
                                                             time.gmtime())
        print "There were %d error(s) and %d warning(s)%s." % \
              (error_count,(param.parameterized.warning_count-initial_warning_count),
               ((" (plus %d warning(s) prior to entering run_batch)"%initial_warning_count
                 if initial_warning_count>0 else "")))
        
        # restore stdout
        sys.stdout = sys.__stdout__
        batch_output.close()
    


def wipe_out_activity():
    """
    Resets activity of all Sheets and their connections to zero.
    """
    # ALERT: this works for now, but it may need to be implemented
    # recursively using methods implemented separately on each class,
    # if there are often new types of objects created that store an
    # activity value.
    for s in topo.sim.objects(Sheet).values():
        s.activity*=0.0
        for c in s.in_connections:
            if hasattr(c,'activity'):
                c.activity*=0.0



def n_bytes():
    """
    Estimate the minimum memory needed for the Sheets in this Simulation, in bytes.

    This estimate is a lower bound only, based primarily on memory for
    the matrices used for activity and connections.
    """
    return sum([s.n_bytes() for s in topo.sim.objects(Sheet).values()])



def n_conns():
    """
    Count the number of connections in all ProjectionSheets in the current Simulation.  
    """     
    return sum([s.n_conns() for s in topo.sim.objects(ProjectionSheet).values()])


def print_sizes():
    """Format the results from n_conns() and n_bytes() for use in batch output."""
    print "Defined %d-connection network; %0.0fMB required for weight storage." % \
    (n_conns(),max(n_bytes()/1024.0/1024.0,1.0))

# added these two function to the PatternDrivenAnalysis hooks 
PatternDrivenAnalysis.pre_presentation_hooks.append(wipe_out_activity)
PatternDrivenAnalysis.pre_presentation_hooks.append(clear_event_queue)

            
# maybe an explicit list would be better?
import types
_public = list(set([_k for _k,_v in locals().items()
                    if isinstance(_v,types.FunctionType) or 
                    (isinstance(_v,type) and issubclass(_v,ParameterizedFunction))
                    and not _v.__name__.startswith('_')]))
_public += [
    "_VersionPrinter",
    "UnpickleEnvironmentCreator",
    "ImportErrorRaisingFakeModule",
    "ImportErrorObject",
]

# Automatically discover all .py files in this directory.
import os,fnmatch
__all__ = _public + [f.split('.py')[0] for f in os.listdir(__path__[0]) if fnmatch.fnmatch(f,'[!._]*.py')]
del f,os,fnmatch



"""
Simple framework to run a set of tests and report the results.

$Id$
"""

import traceback,sys,inspect

# CB: not sure this file should exist, or what it should be called.


def run(tests,title=None):
    """
    Call tests and print the outcomes.

    Each function in tests will be called. Any unhandled exception is
    recorded as a failure; tracebacks from such failures are printed.

    For each test function, prints the name, the docstring, and
    whether it passed or failed.

    Prints and returns the number of errors.
    """
    if title: print "\n%s\n"%title
    errs = {}
    for test in tests:
        test_info = "%s (%s)"%(test.__name__,inspect.getdoc(test))
        error = _run_test(test) 
        if error is None:
            print "PASS: %s"%test_info
        else:
            print "FAIL: %s"%test_info
            errs[(test.__name__,test.__doc__)]=error

    for t,e in errs.items(): print "\n** Error from %s (%s):\n%s\n"%(t[0],t[1],e) 
    print "\nNumber of tests: %s"%len(tests)
    print "Number of errors: %s\n"%len(errs)
    return len(errs)


def _run_test(test):
    """Call test and return None on success, or the exception string on failure."""

    import StringIO
    stderr = StringIO.StringIO()
    sys.stderr = stderr

    # CEBALERT: for some reason, this doesn't always catch errors.
    # I guess somehow they are buried? So in addition to this,
    # we check stderr!
    try:
        test()
        X=None
    except:
        X=traceback.format_exc()
        
    # there's some thing to have bit of code always executed,
    # but can't remember what it is. with it, could make this
    # code cleaner.

    sys.stderr = sys.__stderr__

    if stderr.getvalue(): X = stderr.getvalue()
    stderr.close()
    
    return X

        



"""
Learning functions for Projections.


For example, CFProjectionLearningFunctions compute a new set of
ConnectionFields when given an input and output pattern and a set of
ConnectionField objects.

$Id$
"""

from numpy import ones,zeros
import numpy.oldnumeric as Numeric
from numpy.oldnumeric import Float

import param

from topo.base.cf import CFPLearningFn
from topo.base.sheet import activity_type
from topo.base.functionfamily import Hebbian,LearningFn
# Imported here so that all ProjectionLearningFns will be in the same package
from topo.base.cf import CFPLF_Identity,CFPLF_Plugin  # pyflakes:ignore (API import)


class CFPLF_EuclideanHebbian(CFPLearningFn):
    """
    Hebbian CFProjection learning rule based on Euclidean distance.

    Learning is driven by the distance from the input pattern to the
    weights, scaled by the current activity.  To implement a Kohonen
    SOM algorithm, the activity should be the neighborhood kernel
    centered around the winning unit, as implemented by KernelMax.
    """
    # CEBERRORALERT: ignoring the sheet mask
    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
        # This learning function does not need to scale the learning
        # rate like some do, so it does not use constant_sum_connection_rate()

        cfs = iterator.flatcfs
        rows,cols = output_activity.shape
        for r in xrange(rows):
            for c in xrange(cols):
                flati = r*cols+c
                out = output_activity.flat[flati]
                if out !=0:
                    rate = learning_rate * out                    
                    cf = cfs[flati]
                    X = cf.get_input_matrix(input_activity)
                    cf.weights += rate * (X - cf.weights)

                    # CEBHACKALERT: see ConnectionField.__init__()
                    cf.weights *= cf.mask



#### JABHACKALERT: Untested
##class CFPLF_BCM(CFPLearningFn):
##    """
##    Bienenstock, Cooper, and Munro (1982) learning rule with sliding threshold.
##    
##    (See Dayan and Abbott, 2001, equation 8.12, 8.13).
##
##    Activities change only when there is both pre- and post-synaptic activity.
##    Threshold is adjusted based on recent firing rates.
##    """
##    single_cf_fn = param.ClassSelector(LearningFn,default=BCMFixed())
##    
##    unit_threshold_0=param.Number(default=0.5,bounds=(0,None),
##        doc="Initial value of threshold between LTD and LTP; actual value computed based on recent history.")
##    unit_threshold_learning_rate=param.Number(default=0.1,bounds=(0,None),
##        doc="Amount by which the unit_threshold is adjusted for each activity calculation.")
##
##    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
##        cfs = iterator.proj._cfs
##        # Initialize thresholds the first time we learn the size of the output_activity.
##        if not hasattr(self,'unit_thresholds'):
##            self.unit_thresholds=ones(output_activity.shape,Float32)*self.unit_threshold_0
##
##        rows,cols = output_activity.shape
##
##        # JABALERT: Is this correct?
##      single_connection_learning_rate = self.constant_sum_connection_rate(iterator.proj_n_units,learning_rate)
##
##        # avoid evaluating these references each time in the loop
##        single_cf_fn = self.single_cf_fn
##      for r in xrange(rows):
##            for c in xrange(cols):
##                cf = cfs[r][c]
##                input_act = cf.get_input_matrix(input_activity)
##                unit_activity = output_activity[r,c]
##                threshold=self.unit_thresholds[r,c]
##                #print cf.weights, type(cf.weights)
##                #print input_act, type(input_act)
##                #print single_connection_learning_rate,unit_activity,threshold, (unit_activity-threshold)
##                cf.weights += (single_connection_learning_rate * unit_activity * (unit_activity-threshold)) * input_act 
##                self.unit_thresholds[r,c] += self.unit_threshold_learning_rate*(unit_activity*unit_activity-threshold)
##
##                # CEBHACKALERT: see ConnectionField.__init__()
##                cf.weights *= cf.mask



class CFPLF_Trace(CFPLearningFn):
    """
    LearningFn that incorporates a trace of recent activity,
    not just the current activity.

    Based on P. Foldiak (1991), "Learning Invariance from
    Transformation Sequences", Neural Computation 3:194-200.  Also see
    Sutton and Barto (1981) and Wallis and Rolls (1997).

    Incorporates a decay term to keep the weight vector bounded, and
    so it does not normally require any output_fn normalization for
    stability.
       
    NOT YET TESTED.
    """

    trace_strength=param.Number(default=0.5,bounds=(0.0,1.0),
       doc="How much the learning is dominated by the activity trace, relative to the current value.")     

    single_cf_fn = param.ClassSelector(LearningFn,default=Hebbian(),
        doc="LearningFn that will be applied to each CF individually.")              

    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
        single_connection_learning_rate = self.constant_sum_connection_rate(iterator.proj_n_units,learning_rate)
        ##Initialise traces to zero if they don't already exist
        if not hasattr(self,'traces'):
            self.traces=zeros(output_activity.shape,activity_type)
        for cf,i in iterator():                       
            unit_activity = output_activity.flat[i]
            #   print "unit activity is",unit_activity
        #    print "self trace is",self.traces[r,c]
            new_trace = (self.trace_strength*unit_activity)+((1-self.trace_strength)*self.traces.flat[i])
        #     print "and is now",new_trace
            self.traces.flat[i] = new_trace
            cf.weights += single_connection_learning_rate * new_trace * \
                              (cf.get_input_matrix(input_activity) - cf.weights)
                
            #CEBHACKALERT: see ConnectionField.__init__()
            cf.weights *= cf.mask
      


class CFPLF_OutstarHebbian(CFPLearningFn):
    """
    CFPLearningFunction applying the specified (default is Hebbian) 
    single_cf_fn to each CF, where normalization is done in an outstar-manner.

    Presumably does not need a separate output_fn for normalization.
    
    NOT YET TESTED.
    """
    single_cf_fn = param.ClassSelector(LearningFn,default=Hebbian(),
        doc="LearningFn that will be applied to each CF individually.")

    outstar_wsum = None

    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
        single_connection_learning_rate = self.constant_sum_connection_rate(iterator.proj_n_units,learning_rate)
        # avoid evaluating these references each time in the loop
        single_cf_fn = self.single_cf_fn
        outstar_wsum = zeros(input_activity.shape)
        for cf,i in iterator():
            single_cf_fn(cf.get_input_matrix(input_activity),
                         output_activity.flat[i], cf.weights, single_connection_learning_rate)
            # Outstar normalization
            wrows,wcols = cf.weights.shape
            for wr in xrange(wrows):
                for wc in xrange(wcols):
                    outstar_wsum[wr][wc] += cf.weights[wr][wc]

            # CEBHACKALERT: see ConnectionField.__init__()
            cf.weights *= cf.mask




class HomeoSynaptic(CFPLearningFn):
    """
    Learning function using homeostatic synaptic scaling from 
    Sullivan & de Sa, "Homeostatic Synaptic Scaling in Self-Organizing Maps",
    Neural Networks (2006), 19(6-7):734-43.

    Does not necessarily require output_fn normalization for stability.
    """
    single_cf_fn = param.ClassSelector(LearningFn,default=Hebbian(),
       doc="LearningFn that will be applied to each CF individually")

    beta_n = param.Number(default=0.01,bounds=(0,None),
       doc="homeostatic learning rate")

    beta_c = param.Number(default=0.005,bounds=(0,None),
       doc="time window over which the neuron's firing rate is averaged")

    activity_target = param.Number(default=0.1,bounds=(0,None),
         doc="Target average activity")

    #debug = param.Boolean(default=False,doc="Print average activity values")
    #beta_n = param.Number(default=0.00033,bounds=(0,None),doc="Homeostatic learning rate") #Too small?
    #beta_c = param.Number(default=0.000033,bounds=(0,None),doc="Time window over which the neuron's firing rate is averaged")
    
    def __init__(self,**params):
        super(HomeoSynaptic,self).__init__(**params)
        self.temp_hist = []
        self.ave_hist = []
        
    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
        """
        Update the value of the given weights matrix based on the
        input_activity matrix (of the same size as the weights matrix)
        and the response of this unit (the unit_activity), governed by
        a per-connection learning rate.
        """
        if not hasattr(self,'averages'):
            self.averages = ones(output_activity.shape,Float) * 0.1
            
                            
            # normalize initial weights to 1.0
            for cf,i in iterator():
                current_norm_value = 1.0*Numeric.sum(abs(cf.weights.ravel()))
                if current_norm_value != 0:
                    factor = (1.0/current_norm_value)
                    cf.weights *= factor

        # compute recent average of output activity
        self.averages = self.beta_c * output_activity + (1.0-self.beta_c) * self.averages
        activity_norm = 1.0 + self.beta_n * \
           ((self.averages - self.activity_target)/self.activity_target)

        single_connection_learning_rate = self.constant_sum_connection_rate(iterator.proj_n_units,learning_rate)

        # avoid evaluating these references each time in the loop
        single_cf_fn = self.single_cf_fn
        for cf,i in iterator():
            single_cf_fn(cf.get_input_matrix(input_activity),
                         output_activity.flat[i], cf.weights, single_connection_learning_rate)

            # homeostatic normalization
            cf.weights /= activity_norm.flat[i]

            # CEBHACKALERT: see ConnectionField.__init__()
            cf.weights *= cf.mask
         
        # For analysis only; can be removed (in which case also remove the initializations above)
# CEBALERT: I changed [0][7] to [0]!
        self.ave_hist.append(self.averages.flat[0])
        self.temp_hist.append (Numeric.sum(abs(iterator.flatcfs[0].weights.ravel())))




class CFPLF_PluginScaled(CFPLearningFn):
    """
    CFPLearningFunction applying the specified single_cf_fn to each CF.
    Scales the single-connection learning rate by a scaling factor
    that is different for each individual unit. Thus each individual
    connection field uses a different learning rate.
    """

    single_cf_fn = param.ClassSelector(LearningFn,default=Hebbian(),
        doc="Accepts a LearningFn that will be applied to each CF individually.")

    learning_rate_scaling_factor = param.Parameter(default=None,
        doc="Matrix of scaling factors for scaling the learning rate of each CF individually.")

    
    def __call__(self, iterator, input_activity, output_activity, learning_rate, **params):
        """Apply the specified single_cf_fn to every CF."""
       
        if self.learning_rate_scaling_factor is None:
            self.learning_rate_scaling_factor = ones(output_activity.shape)
            
        single_cf_fn = self.single_cf_fn
        single_connection_learning_rate = self.constant_sum_connection_rate(iterator.proj_n_units,learning_rate)
        
        for cf,i in iterator():
            sc_learning_rate = self.learning_rate_scaling_factor.flat[i] * single_connection_learning_rate 
            single_cf_fn(cf.get_input_matrix(input_activity),
                         output_activity.flat[i], cf.weights, sc_learning_rate)
            # CEBHACKALERT: see ConnectionField.__init__() re. mask & output fn
            cf.weights *= cf.mask   
      

    def update_scaling_factor(self, new_scaling_factor):
        """Update the single-connection learning rate scaling factor."""
        self.learning_rate_scaling_factor = new_scaling_factor


__all__ = [
    "CFPLF_Identity",
    "CFPLF_Plugin",
    "CFPLF_EuclideanHebbian",
    "CFPLF_Trace",
    "CFPLF_OutstarHebbian",
    "HomeoSynaptic",
    "CFPLF_PluginScaled",
]



"""
General utility functions and classes.

$Id$
"""

import re
import random
import numpy
import functools



def NxN(tuple):
    """
    Converts a tuple (X1,X2,...,Xn) to a string 'X1xX2x...xXn'
    """    
    return 'x'.join([`N` for N in tuple])



enum = enumerate


class Struct:
    """
    A simple structure class, taking keyword args and assigning them to attributes.

    For instance:
    
    s = Struct(foo='a',bar=1)
    >>> s.foo
    'a'
    >>> s.bar
    1


    From http://www.norvig.com/python-iaq.html
    """
    def __init__(self, **entries): self.__dict__.update(entries)

    def __repr__(self):
        # 
        args = ['%s=%s' % (k, repr(v)) for (k,v) in vars(self).items()]
        return 'Struct(%s)' % ', '.join(args)



def flat_indices(shape):
    """
    Returns a list of the indices needed to address or loop over all
    the elements of a 1D or 2D matrix with the given shape. E.g.:

    flat_indices((3,)) == [0,1,2]
    """
    if len(shape) == 1:
        return [(x,) for x in range(shape[0])]
    else:
        rows,cols = shape
        return [(r,c) for r in range(rows) for c in range(cols)]
        
    

def flatten(l):
    """
    Flattens a list.
    
    Written by Bearophile as published on the www.python.org newsgroups.
    Pulled into Topographica 3/5/2005.
    """
    if type(l) != list:
        return l
    else:
        result = []
        stack = []
        stack.append((l,0))
        while len(stack) != 0:
            sequence, j = stack.pop(-1)
            while j < len(sequence):
                if type(sequence[j]) != list:
                    k, j, lens = j, j+1, len(sequence)
                    while j < len(sequence) and \
                          (type(sequence[j]) != list):
                        j += 1
                    result.extend(sequence[k:j])
                else:
                    stack.append((sequence, j+1))
                    sequence, j = sequence[j], 0
        return result

    

"""
Return the cross-product of a variable number of lists (e.g. of a list of lists).

Use to obtain permutations, e.g.
l1=[a,b]
l2=[c,d]
cross_product([l1,l2]) = 
[[a,c], [a,d], [b,c], [b,d]]


From:
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/159975
"""
# Need to re-write so someone other than Python knows what might happen when this runs
cross_product=lambda ss,row=[],level=0: len(ss)>1 \
   and reduce(lambda x,y:x+y,[cross_product(ss[1:],row+[i],level+1) for i in ss[0]]) \
   or [row+[i] for i in ss[0]]



# JABALERT: Should frange be replaced with numpy.arange or numpy.linspace?
def frange(start, end=None, inc=1.0, inclusive=False):
    """
    A range function that accepts float increments.

    Otherwise, works just as the inbuilt range() function.  If
    inclusive is False, as in the default, the range is exclusive (not
    including the end value), as in the inbuilt range(). If inclusive
    is true, the range may include the end value.
        
    'All theoretic restrictions apply, but in practice this is
    more useful than in theory.'

    From: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66472
    """
    if end == None:
        end = start + 0.0
        start = 0.0

    # Increments of zero would lead to an infinite loop, which can happen if
    # this is mistakenly called with a integer-based rational expression like 1/2.
    assert ((inc>0 and start<=end) or (inc<0 and start>=end))
    
    L = []
    while 1:
        next = start + len(L) * inc
        if inclusive:
          if inc > 0 and next > end: break
          elif inc < 0 and next < end: break
        else:
          if inc > 0 and next >= end: break
          elif inc < 0 and next <= end: break
        L.append(next)
        
    return L



def shortclassname(x):
    """
    Returns the class name of x as a string with the leading package information removed.

    E.g. if x is of type "<class 'topo.base.sheet.Sheet'>", returns
    "Sheet"
    """
    return re.sub("'>","",re.sub(".*[.]","",repr(type(x))))



def profile(command,n=50,sorting=('cumulative','time'),strip_dirs=False):
    """
    Profile the given command (supplied as a string), printing
    statistics about the top n functions when ordered according to
    sorting.

    sorting defaults to ordering by cumulative time and then internal
    time; see http://docs.python.org/lib/profile-stats.html for other
    sorting options.

    By default, the complete paths of files are not shown. If there
    are multiple files with the same name, you might wish to set
    strip_dirs=False to make it easier to follow the output.


    Examples:

    - profile loading a simulation:
    profile('execfile("examples/hierarchical.ty")')

    - profile running an already loaded simulation:
    profile('topo.sim.run(10)')

    - profile running a whole simulation:
    profile('execfile("examples/lissom_oo_or.ty");topo.sim.run(20000)')

    - profile running a simulation, but from the commandline:
    ./topographica examples/hierarchical.ty -c "from topo.misc.util import profile; profile('topo.sim.run(10)')"
    """
    # This function simply wraps some functions from the cProfile
    # module, making profiling easier.
    import cProfile, pstats


    # CB: leaves around "filename": should give this a proper name and maybe
    # put in /tmp/ and maybe allow someone to choose where to save it
    cProfile.run(command,'filename')
    prof_stats = pstats.Stats('filename')

    if strip_dirs:prof_stats.strip_dirs()
    
    prof_stats.sort_stats(*sorting).print_callees(n)
    ### the above lets us see which times are due to which calls
    ### unambiguously, while the version below only reports total time
    ### spent in each object, not the time due to that particular
    ### call.
    prof_stats.sort_stats(*sorting).print_stats(n)

    


def weighted_sample(seq,weights=[]):
    """
    Select randomly from the given sequence.
    
    The weights, if given, should be a sequence the same length as
    seq, as would be passed to weighted_sample_idx().
    """
    if not weights:
        return seq[random.randrange(len(seq))]
    else:
        assert len(weights) == len(seq)
        return seq[weighted_sample_idx(weights)]


def weighted_sample_idx(weights):
    """
    Return an integer generated by sampling the discrete distribution
    represented by the sequence of weights.

    The integer will be in the range [0,len(weights)).  The weights
    need not sum to unity, but can contain any non-negative values
    (e.g., [1 1 1 100] is a valid set of weights).

    To use weights from a 2D numpy array w, specify w.ravel() (not the
    w.flat iterator).
    """
    total = sum(weights)
    if total == 0:
        return random.randrange(len(weights))
    index = random.random() * total
    accum = 0
    for i,x in enumerate(weights):
        accum += x
        if index < accum:
            return i


def idx2rowcol(idx,shape):
    """
    Given a flat matrix index and a 2D matrix shape, return the (row,col)
    coordinates of the index.
    """
    assert len(shape) == 2
    rows,cols = shape

    return idx/cols,idx%cols



def rowcol2idx(r,c,shape):
    """
    Given a row, column, and matrix shape, return the corresponding index
    into the flattened (raveled) matrix.
    """
    assert len(shape) == 2
    rows,cols = shape

    return r * cols + c


    
def centroid(pts,weights):
    """
    Return the centroid of a weighted set of points as an array.

    The pts argument should be an array of points, one per row,
    and weights should be a vector of weights.
    """
    # CEBALERT: use numpy.sum? Worthwhile if weights is a numpy.array.
    return numpy.dot(numpy.transpose(pts),weights)/sum(weights)



def signabs(x):
    """
    Split x into its sign and absolute value.

    Returns a tuple (sign(x),abs(x)).  Note: sign(0) = 1, unlike
    numpy.sign.
    """

    if x < 0:
        sgn = -1
    else:
        sgn = 1

    return sgn,abs(x)


def linearly_interpolate(table,value):
    """
    Interpolate an appropriate value from the given list of values, by number.

    Assumes the table is a list of items to be returned for integer values, 
    and interpolates between those items for non-integer values.
    """ 

    lower_index=int(value)
    upper_index=lower_index+1

    # Intermediate value; interpolate or return exact value as appropriate
    if lower_index+1<len(table):
       lookup=table[lower_index]+(value%1.0)*(table[upper_index]-table[lower_index])
    
    # Upper bound -- return largest value
    elif lower_index+1==len(table):
       lookup=table[len(table)-1]

    # Over upper bound -- return largest value and print warning
    # JABALERT: Printing a warning message is not necessarily the most
    # useful behavior. Should at least provide some identification of
    # where the warning is coming from. Should add bounds_error and/or
    # bounds_warn options. (Could turn into a ParameterizedFunction so
    # that any message includes an identification, and so that
    # warnings-as-errors will work.) Would be nice if we could use the
    # equivalent function from scipy, but we can't yet depend on scipy
    # being available.
    else:
       lookup=table[len(table)-1]
       print "Warning -- value %f out of range; returning maximum of %f" % (value,lookup)

    return lookup



# CB: note that this has only really been tested for output;
# I've never tried using it to e.g. read multiple files.
class MultiFile(object):
    """
    For all file_like_objs passed on initialization, provides a
    convenient way to call any of file's methods (on all of them).

    E.g. The following would cause 'test' to be written into two
    files, as well as to stdout:

    import sys
    f1 = open('file1','w')
    f2 = open('file2','w')
    m = MultiFile(f1,f2,sys.stdout)
    m.write('test')
    """
    def __init__(self,*file_like_objs):
        self.file_like_objs=file_like_objs
        self.__provide_file_methods()

    def __provide_file_methods(self):
        # Provide a version of all the methods of the type file that
        # don't start with '_'. In each case, the provided version
        # loops over all the file_like_objs, calling the file method
        # on all of them.
        file_methods = [attr for attr in file.__dict__
                        if not attr.startswith('_')
                        and callable(file.__dict__[attr])]

        for method in file_methods:
            def looped_method(method_,*args,**kw):
                all_out = []
                for output in self.file_like_objs:
                    out = getattr(output,method_)(*args,**kw)
                    all_out.append(out)
                return all_out

            setattr(self,method,functools.partial(looped_method,method))




# CEBALERT: should be moved to legacy, once that file is reorganized
# so that classes can be imported from it relatively independently,
# without requiring all the legacy functions to be loaded.
# (Right now, topo.__init__ can't import gmpyFaker from
# topo.misc.legacy because importing topo.misc.legacy causes various
# pieces of code to run that depend on topo.sim existing.)

############################################################        
# Alternative module faking using import hooks (see
# http://www.python.org/dev/peps/pep-0302/).
# Based on http://orestis.gr/blog/2008/12/20/python-import-hooks/.
import sys,imp
class ModuleFaker(object):
    def load_module(self,name):
        if name not in sys.modules:
            module = self.create_module(name)
            module.__file__ = self.path
            sys.modules[name] = module
            if '.' in name:
                parent_name, child_name = name.rsplit('.', 1)
                setattr(sys.modules[parent_name], child_name, module)
        return sys.modules[name]
    
    def create_module(self,name):
        raise NotImplementedError

class ModuleImporter(object):
    def find_module(self,fullname,path=None):
        raise NotImplementedError

class gmpyFaker(ModuleFaker):
    def create_module(self,name):
        module = imp.new_module(name)
        # CEBALERT: not sure what precision should be used for FixedPoint to
        # replace rational. Should we set the precision really high?
        code = \
"""
from __future__ import division
import topo.misc.fixedpoint as fixedpoint
import param
class mpq(object):
    def __new__(self,*args,**kw):
        n = fixedpoint.FixedPoint(eval(str(args[0])),precision=4)
        param.Parameterized().warning("gmpy.mpq('%s') replaced by fixedpoint.FixedPoint('%s')"%(args[0],n))
        return n
"""
        exec code in module.__dict__
        return module

class gmpyImporter(ModuleImporter):

    def find_module(self, fullname, path=None):
        if fullname == 'gmpy' or fullname.startswith('gmpy.'):
            import param
            param.Parameterized().warning('Module "gmpy" is not available. gmpy.mpq is provided by using fixedpoint.FixedPoint.')
            g = gmpyFaker()
            g.path = path
            return g
        return None



"""
Object classes for recording and plotting time-series data.

This module defines a set of DataRecorder object types for recording
time-series data, a set of Trace object types for
specifying ways of generating 1D-vs-time traces from recorded data,
and a TraceGroup object that will plot a set of traces on stacked,
aligned axes.

$Id$
"""

import os
import bisect
from itertools import izip

from numpy import asarray
import ImageDraw

import param
from param import normalize_path

from topo.base.simulation import EventProcessor
from topo.plotting.bitmap import RGBBitmap, MontageBitmap, TITLE_FONT
from topo.misc.util import Struct



class DataRecorder(EventProcessor):
    """
    Record time-series data from a simulation.

    A DataRecorder instance stores a set of named time-series
    variables, consisting of a sequence of recorded data items of any
    type, along with the times at which they were recorded.

    DataRecorder is an abstract class for which different
    implementations may exist for different means of storing recorded
    data.  For example, the subclass InMemoryRecorder stores all the
    data in memory.

    A DataRecorder instance can operate either as an event processor, or in a
    stand-alone mode.  Both usage modes can be used on the same
    instance in the same simulation.

    STAND-ALONE USAGE:

    A DataRecorder instance is used as follows:
    
      - Method .add_variable adds a named time series variable.
      - Method .record_data records a new data item and timestamp.
      - Method .get_data gets a time-delimited sequence of data from a variable
    
    EVENTPROCESSOR USAGE:

    A DataRecorder can also be connected to a simulation as an event
    processor, forming a kind of virtual recording equipment.  An
    output port from any event processor in a simulation can be
    connected to a DataRecorder; the recorder will automaticall create
    a variable with the same name as the connection, and record any
    incoming data on that variable with the time it was received.  For
    example::

      topo.sim['Recorder'] =  InMemoryRecorder()
      topo.sim.connect('V1','Recorder',name='V1 Activity')

    This script snippet will create a new DataRecorder and
    automatically record all activity sent from the sheet 'V1'.
    """

    __abstract = True


    def __init__(self,**params):
        super(DataRecorder,self).__init__(**params)
        self._trace_groups = {}


    def _src_connect(self,conn):
        raise NotImplementedError

    
    def _dest_connect(self,conn):
        super(DataRecorder,self)._dest_connect(conn)
        self.add_variable(conn.name)


    def input_event(self,conn,data):
        self.record_data(conn.name,self.simulation.time(),data)


    def add_variable(self,name):
        """
        Create a new time-series variable with the given name.
        """
        raise NotImplementedError


    def record_data(self,varname,time,data):
        """
        Record the given data item with the given timestamp in the
        named timeseries.
        """
        raise NotImplementedError


    def get_data(self,varname,times=(None,None),fill_range=False):
        """
        Get the named timeseries between the given times
        (inclusive). If fill_range is true, the returned data will
        have timepoints exactly at the start and end of
        the given timerange.  The data values at these timepoints will
        be those of the next-earlier datapoint in the series.
        
        (NOTE: fill_range can fail to create a beginning timepoint if
        the start of the time range is earlier than the first recorded datapoint.]
        """
        raise NotImplementedError


    def get_times(self,var):
        """
        Get all the timestamps for a given variable.
        """
        raise NotImplementedError


    def get_time_indices(self,varname,start_time,end_time):
        """
        For the named variable, get the start and end indices suitable
        for slicing the data to include all times t::
        
          start_time <= t <= end_time.

        A start_ or end_time of None is interpreted to mean the
        earliest or latest available time, respectively.
        """

        times = self.get_times(varname)
        if start_time is None:
            start = 0
        else:
            start = bisect.bisect_left(times,start_time)
            if start >= len(times):
                start = len(times)-1
            elif times[start] > start_time:
                start -= 1

        if end_time is None:
            end = None
        else:
            end = bisect.bisect_right(times,end_time)

        return start,end


                    

class InMemoryRecorder(DataRecorder):
    """
    A data recorder that stores all recorded data in memory.
    """

    def __init__(self,**params):
        super(InMemoryRecorder,self).__init__(**params)
        self._vars = {}


    def add_variable(self,name):
        self._vars[name] = Struct(time=[],data=[])


    def record_data(self,varname,time,data):
        var = self._vars[varname]

        # add the data, maintaining it sorted by time
        if not var.time or var.time[-1] <= time:
            var.time.append(time)
            var.data.append(data)
        elif time < var.time[0]:
            var.time.insert(0,time)
            var.data.insert(0,data)
        else:
            idx = bisect.bisect_right(var.time,time)
            var.time.insert(idx,time)
            var.data.insert(idx,data)

    def get_datum(self,name,time):
        idx,dummy = self.get_time_indices(name,time,time)
        data = self._vars[name].data
        if idx >= len(data):
            idx -= 1
        return data[idx]
    
    def get_data(self,name,times=(None,None),fill_range=False):
        tstart,tend = times
        start,end = self.get_time_indices(name,tstart,tend)
        var = self._vars[name]

        if start >= len(var.data):
            # if the start index is out of bounds
            if fill_range:
                time = times
                data = [var.data[-1]]*2
            else:
                time,data = [],[]
        else:
            time,data = var.time[start:end],var.data[start:end]
            if fill_range:
                if time[0] > tstart and start > 0:
                    time.insert(0,tstart)
                    data.insert(0,var.data[start-1])
                if time[-1] < tend:
                    time.append(tend)
                    data.append(data[-1])
        
        return time,data


    def get_times(self,varname):
        return self._vars[varname].time




class Trace(param.Parameterized):
    """
    A specification for generating 1D traces of data from recorded
    timeseries.

    A Trace object is a callable object that encapsulates
    a method for generating a 1-dimensional trace from possibly
    multidimensional timeseries data, along with a specification for
    how to plot that data, including Y-axis boundaries and plotting arguments.
    
    Trace is an abstract class.  Subclasses implement
    the __call__ method to define how to extract a 1D trace from a
    sequence of data.
    """
    
    __abstract = True

    data_name = param.String(default=None,doc="""
        Name of the timeseries from which the trace is generated.
        E.g. the connection name into a DataRecorder object.""")

    # JPALERT: This should really be something like a NumericTuple,
    # except that NumericTuple won't allow the use of None to indicate
    # 'no default'.  (Nor will Number.)
    # JB: We could have that as an option, or as another Parameter
    # type, but in many cases knowing that the parameter cannot be set
    # to a non-numeric value is crucial, as it means we don't have to
    # do special checks every time the value is used.  So we should
    # leave the default behavior as it is, but yes, it would be good
    # to handle None for numeric types (and also for Boolean, to make
    # it tri-state).  Note that Bounds is a specific type that we
    # should probably support in any case, because it not only needs
    # to support None, it needs to specify whether the bounds are
    # inclusive or exclusive.
    ybounds = param.Parameter(default=(None,None),doc="""
        The (min,max) boundaries for y axis.  If either is None, then
        the bound min or max of the data given, respectively.""")

    ymargin = param.Number(default=0.1,doc="""
        The fraction of the difference ymax-ymin to add to the
        top of the plot as padding.""")
    
    plotkw = param.Dict(default=dict(linestyle='steps'),doc="""
        Contains the keyword arguments to pass to the plot command
        when plotting the trace.""")

       
    def __call__(self,data):
        raise NotImplementedError


    # JB: Needs docstring.  Should this be a property instead?
    def get_ybounds(self,ydata):
        ymin,ymax = self.ybounds
        if ymax is None:
            ymax = max(ydata)

        if ymin is None:
            ymin = min(ydata)

        ymax += (ymax-ymin)*self.ymargin

        return ymin,ymax


class IdentityTrace(Trace):
    """
    A Trace that returns the data, unmodified.
    """
    def __call__(self,data):
        return data



class IndexTrace(Trace):
    """
    A Trace that assumes that each data item is a sequence that can be
    indexed with a single integer, and traces the value of one indexed element.
    """
    
    index = param.Integer(default=0,doc="""
        The index into the data to be traced.""")
    
    def __call__(self,data):
        return [x[self.index] for x in data]



class SheetPositionTrace(Trace):
    """
    A trace that assumes that the data are sheet activity matrices,
    and traces the value of a given (x,y) position on the sheet.
    """
    
    x = param.Number(default=0.0,doc="""
        The x sheet-coordinate of the position to be traced.""")
    
    y = param.Number(default=0.0,doc="""
        The y sheet-coordinate of the position to be traced.""")

    position = param.Composite(attribs=['x','y'],doc="""
        The sheet coordinates of the position to be traced.""")

    # JPALERT:  Would be nice to some way to set up the coordinate system
    # automatically.  The DataRecorder object already knows what Sheet
    # the data came from.
    coordframe = param.Parameter(default=None,doc="""
        The SheetCoordinateSystem to use to convert the position
        into matrix coordinates.""")

    def __call__(self,data):
        r,c = self.coordframe.sheet2matrixidx(self.x,self.y)
        return [d[r,c] for d in data]



class TraceGroup(param.Parameterized):
    """
    A group of data traces to be plotted together.

    A TraceGroup defines a set of associated data traces and allows
    them to be plotted on stacked, aligned axes.  The constructor
    takes a DataRecorder object as a data source, and a list of
    Trace objects that indicate the traces to plot.  The
    trace specifications are stored in the attribute self.traces,
    which can be modified at any time.
    """


    hspace = param.Number(default=0.6,doc="""
       Height spacing adjustment between plots.  Larger values
       produce more space.""")

    time_axis_relative = param.Boolean(default=False,doc="""
       Whether to plot the time-axis tic values relative to the start
       of the plotted time range, or in absolute values.""")

    
    def __init__(self,recorder,traces=[],**params):
        super(TraceGroup,self).__init__(**params)
        self.traces = traces
        self.recorder = recorder


    def plot(self,times=(None,None)):
        """
        Plot the traces.

        Requires MatPlotLib (aka pylab).
        
        Plots the traces specified in self.traces, over the timespan
        specified by times.  times = (start_time,end_time); if either
        start_time or end_time is None, it is assumed to extend to the
        beginning or end of the timeseries, respectively.
        """
        
        import pylab
        rows = len(self.traces)
        tstart,tend = times
        
        pylab.subplots_adjust(hspace=self.hspace)
        for i,trace in enumerate(self.traces):
            # JPALERT: The TraceGroup object should really create its
            # own matplotlib.Figure object and always plot there
            # (instead of in the frontmost plot), but I haven't
            # figured out how to do that yet.
            pylab.subplot(rows,1,i+1)
            pylab.title(trace.name)
            time,data = self.recorder.get_data(trace.data_name,times=times,fill_range=True)
            y = trace(data)
            if self.time_axis_relative:
                time = asarray(time) - time[0]
            pylab.plot(time,y,**trace.plotkw)
            ymin,ymax = trace.get_ybounds(y)
            pylab.axis(xmin=time[0],xmax=time[-1],ymin=ymin,ymax=ymax)        
            




def get_images(name,times,recorder,overlays=(0,0,0)):
    """
    Get a time-sequence of matrix data from a DataRecorder variable
    and convert it to a sequence of images stored in Bitmap objects.

    Parameters: name is the name of the variable to be queried. times
    is a sequence of timepoints at which to query the
    variable. recorder is the data recorder. overlays is a tuple of
    matrices or scalars to be added to the red, green, and blue
    channels of the bitmaps respectively.
    """
    result = []

    for t in times:
        d = recorder.get_datum(name,t)
        im = RGBBitmap(d+overlays[0],d+overlays[1],d+overlays[2])
        result.append(im)
    return result


# JABALERT: Is there some reason it is called ActivityMovie in
# particular, if it can plot things other than Activity?
# Maybe DataRecorderMovie?
class ActivityMovie(param.Parameterized):
    """
    An object encapsulating a series of movie frames displaying the
    value of one or more matrix-valued time-series contained in a
    DataRecorder object.

    An ActivityMovie takes a DataRecorder object, a list of names of
    variables in that recorder and a sequence of timepoints at which
    to sample those variables.  It uses that information to compose a
    sequence of MontageBitmap objects displaying the stored values of
    each variable at each timepoint.  These bitmaps can then be saved
    to sequentially-named files that can be composited into a movie by
    external software.

    Parameters are available to control the layout of the montage,
    adding timecodes to the frames, and the names of the frame files.
    """

    
    variables = param.List(class_=str, doc="""
        A list of variable names in a DataRecorder object containing
        matrix-valued time series data.""")
       
    overlays = param.Dict(default={}, doc="""
        A dictionary indicating overlays for the variable bitmaps.  The
        for each key in the dict matching the name of a variable, there
        should be associated a triple of matrices to be overlayed on
        the red, green, and blue channels of the corresponding bitmap
        in each frame.""")
       
    frame_times = param.List(default=[0,1], doc="""
        A list of the times of the frames in the movie.""")
       
    montage_params = param.Dict(default={},doc="""
        A dictionary containing parameters to be used when
        instantiating the MontageBitmap objects representing each frame.""",       
        instantiate=False)

    recorder = param.ClassSelector(class_=DataRecorder, doc="""
        The DataRecorder storing the timeseries.""")

    filename_fmt = param.String(default='%n_%t.%T',doc="""
        The format for the filenames used to store the frames.  The following
        substitutions are possible:
        
        %n: The name of this ActivityMovie object.
        %t: The frame time, as formatted by the filename_time_fmt parameter
        %T: The filetype given by the filetype parameter. """)
    
    filename_time_fmt = param.String(default='%05.0f', doc="""
        The format of the frame time, using Python string substitution for
        a floating-point number.""")
       
    filetype = param.String(default='tif',doc="""
        The filetype to use when writing frames. Can be any filetype understood
        by the Python Imaging Library.""")

    filename_prefix = param.String(default='', doc="""
        A prefix to prepend to the filename of each frame when saving; 
        can include directories.  If the filename contains a path, any
        non-existent directories in the path will be created when the
        movie is saved.""")

    add_timecode = param.Boolean(default=False, doc="""
        Whether to add a visible timecode indicator to each frame.""")

    timecode_options = param.Dict(default={},instantiate=False,doc="""
        A dictionary of keyword options to be passed to the PIL ImageDraw.text method
        when drawing the timecode on the frame. Valid options include font,
        an ImageFont object indicating the text font, and fill a PIL color
        specification indicating the text color.  If unspecified, color defaults to
        the PIL default of black.  Font defaults to topo.plotting.bitmap.TITLE_FONT.""")
    
    timecode_fmt = param.String(default='%05.0f',doc="""
        The format of the timecode displayed in the movie frames, using
        Python string substitution for a floating-point number.""")
    
    timecode_offset = param.Number(default=0,doc="""
        A value to be added to each timecode before formatting for display.""")
    

    def __init__(self,**params):
        super(ActivityMovie,self).__init__(**params)

        bitmaps = [get_images(var,self.frame_times,self.recorder,
                              overlays=self.overlays.get(var,(0,0,0)))
                   for var in self.variables]
        
        self.frames = [MontageBitmap(bitmaps=list(bms),**self.montage_params)
                       for bms in izip(*bitmaps)]
        if self.add_timecode:
            for t,f in izip(self.frame_times,self.frames):
                draw = ImageDraw.Draw(f.image)
                timecode = self.timecode_fmt % (t+self.timecode_offset)
                tw,th = draw.textsize(timecode,font=self.timecode_options.setdefault('font',TITLE_FONT))
                w,h = f.image.size
                
                draw.text((w-tw-f.margin-1,h-th-1),timecode,**self.timecode_options)

    
    def save(self):
        """Save the movie frames."""
        
        filename_pat = self.name.join(self.filename_fmt.split('%n'))
        filename_pat = self.filename_time_fmt.join(filename_pat.split('%t'))
        filename_pat = self.filetype.join(filename_pat.split('%T'))

        filename_pat = normalize_path(filename_pat,prefix=self.filename_prefix)
        dirname = os.path.dirname(filename_pat)
        if not os.access(dirname,os.F_OK):
            os.makedirs(dirname)
        
        self.verbose('Writing',len(self.frames),'to files like "%s"'%filename_pat)
        for t,f in zip(self.frame_times,self.frames):
            filename = filename_pat% t
            self.debug("Writing frame",repr(filename))
            f.image.save(filename)



"""
File saving routines for plots.

Typically called using save_plotgroup in commands/analysis.py, but these
objects can also be instantiated explicitly, to save a series of plots.

$Id$
"""

import Image
import numpy

import param
from param import normalize_path

import topo


# Consider using PIL's ImageFont module

class PlotGroupSaver(param.Parameterized):
    """
    Allows a PlotGroup to be saved as a set of bitmap files on disk.
    """
    file_format = param.String(default="png",doc="""
        Bitmap image file format to use.""")

    filename_prefix = param.String(default="",doc="""
        Optional prefix that can be used in the filename_format command
        to disambiguate different simulations or conditions.""")

    filename_suffix = param.String(default="",doc="""
        Optional suffix that can be used in the filename_format command
        to disambiguate different simulations or conditions.""")

    filename_format = param.String(default=
        "%(filename_prefix)s%(basename)s_%(plot_label)s%(filename_suffix)s.%(file_format)s",doc="""
        Format string to use for generating filenames for plots.  This
        string will be evaluated in the context of a dictionary that
        defines various items commonly used when generating filenames,
        including::

          basename:    the default sim.basename(), usually name+time()
          time:        the current simulation time (topo.sim.time())
          sim_name:    the name of the current simulation (topo.sim.name)
          plot_label:  the label specfied in the PlotGroup for this plot
          file_format: the bitmap image file format for this type of plot
          plotgroup_name: the name of this PlotGroup
      """)
    # Should move this out of plotfilesaver to get the same filenames in the GUI.
    # Should also allow each template in topo/command/analysis.py to have a nice
    # short filename format, perhaps as an option.

    def __init__(self,plotgroup,**params):
        super(PlotGroupSaver,self).__init__(**params)
        self.plotgroup = plotgroup 


    def strip(self,filename):
        """Strip inappropriate characters from a filename."""
        return filename\
            .replace('\n','_')\
            .replace('\t','_')\
            .replace(' ','_')\
            .replace('&','And')


    def filename(self,label,**params):
        """Calculate a specific filename from the filename_format."""
        varmap = dict(self.get_param_values())
        varmap.update(self.__dict__)
        varmap['basename']= topo.sim.basename()
        varmap['sim_name']= topo.sim.name
        varmap['time']=topo.sim.time()
        varmap['plot_label']=label
        varmap['plotgroup_name']=self.plotgroup.name
        varmap.update(params)
        
        return self.strip(self.filename_format % varmap)


    def save_to_disk(self,**params):
        for p,l in zip(self.plotgroup.plots,self.plotgroup.labels):
            p.bitmap.image.save(normalize_path(self.filename(l,**params)))






# Could move this elsewhere if it will be useful.
#
# Adapted from:
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/412982
def make_contact_sheet(imgs, (marl,mart,marr,marb), padding):
    """
    Make a contact sheet (image grid) from a 2D array of PIL images::
    
      imgs     2D array of images
               
      marl     The left margin in pixels
      mart     The top margin in pixels
      marr     The right margin in pixels
      marb     The bottom margin in pixels

      padding  The padding between images in pixels

    Returns a PIL image object.
    """
    # should make sure imgs is numpy array
    
    # CB: *** should do this in numpy without the conversion to list and back! ***
    nrows,ncols = imgs.shape
    i_widths  = numpy.array([i.size[0] for i in imgs.ravel()]).reshape(nrows,ncols)
    i_heights = numpy.array([i.size[1] for i in imgs.ravel()]).reshape(nrows,ncols)

    col_widths = i_widths.max(axis=0)
    row_heights = i_heights.max(axis=1)
    
    marw = marl+marr
    marh = mart+ marb

    padw = (ncols-1)*padding
    padh = (nrows-1)*padding

    isize = (col_widths.sum()+marw+padw, row_heights.sum()+marh+padh)

    # Create the new image. The background doesn't have to be white.
    white = (255,255,255)
    inew = Image.new('RGB',isize,white)

    # CB: should be replaced with a more numpy technique.        
    for irow in range(nrows):
        for icol in range(ncols):
            # if different widths in a col, or different heights in a
            # row, each image will currently just go at top left
            # defined by largest in col/row (should be centered
            # instead)
            left = marl+col_widths[0:icol].sum()+icol*padding
            right = left+i_widths[irow,icol]
            upper = mart+row_heights[0:irow].sum()+irow*padding
            lower = upper+i_heights[irow,icol] 
            inew.paste(imgs[irow,icol],(left,upper,right,lower))
    return inew




class CFProjectionPlotGroupSaver(PlotGroupSaver):
    """
    Allows a CFProjectionPlotGroup to be saved as a bitmap file,
    concatenating all the CF plots into a single image.
    """
    def save_to_disk(self,**params):
        imgs = numpy.array([p.bitmap.image
                            for p in self.plotgroup.plots],
                           dtype=object).reshape(
            self.plotgroup.proj_plotting_shape)
        img = make_contact_sheet(imgs, (3,3,3,3), 3)
        img.save(normalize_path(self.filename(
            self.plotgroup.sheet.name+"_"+
            self.plotgroup.projection.name,**params)))



"""

$Id$
"""

# CEBHACKALERT: we should delete this file, right?

import unittest

from numpy.oldnumeric import array, Float32

# Currently empty; should be expanded
cases = []

suite = unittest.TestSuite()
suite.addTests(unittest.makeSuite(case) for case in cases)
              
if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=1).run(suite)
    


        
